"use strict";

var xmlparse, request, isTi = false;

// work in Titanium
if (typeof Ti === "undefined") {
   xmlparse = require('jsdom').env;
   request  = require('request').get;
} else {
   isTi = true;
   xmlparse = Ti.XML.parseString;
   request = function (url, callback) {
      var xhr = Ti.Network.createHTTPClient();
      xhr.onload = function (e) {
         var response = {
            statusCode  : e.source.status
         };
         callback(null, response, e.source.responseText);
      };
      xhr.onerror = function (e) {
         var response = {
            statusCode  : e.source.status
         };
         callback(new Error("response: " + e.source.status), response, null);
      };
   };
}

/*
   Class: food
   Allows easy querying of the Rutgers Dining Public XML feed for meal data.
*/

exports.client = function () {
   var exports = {},
       baseURL,
       locations,
       mealNames;

   baseURL = "http://dining1.rutgers.edu/FoodPro/menu.asp?";
   locations = {
      busch: {
         title: 'Busch Dining Hall',
         id: '04'
      },
      brower: {
         title: 'Brower Commons',
         id: '01'
      },
      neilson: {
         title: 'Neilson Dining Hall',
         id: '05'
      },
      livingston: {
         title: 'Livingston Dining Commons',
         id: '03'
      }
   };

   /* Group: Public functions
    
    Function: getMenu
    Takes in a location and a callback and calls yo callback function with 
    a menu object that has data about the food being served at that location. 

    Parameters:
         location    - *string* One of 'busch', 'brower', 'neilson', 
                       'livingston'
         mealName    - *string* One of "breakfast", "lunch", "dinner",
                       'knight room'
         callback    - *function(err, menu)* called with results

    Returns:
      Nothing. The callback is what you should use.

    Example:
    (start code)
    food.getMenu("busch", "breakfast", function(err, menu) {
       console.log(menu);
    });

    { 'Breakfast Meats': 
         [ 'Canadian Bacon',
          'Grilled Turkey Sausage Links',
          'Vegetarian Sausage Patties' ],
        'Breakfast Entrees': 
         [ 'Belgian  Waffles',
          'Hard Boiled Eggs',
          'Old Fashioned French Toast',
          'Scramble Cholesterol Free Egg',
          'Scrambled Eggs' ],
        'Breakfast Bakery': 
         [ 'Assorted Bagels',
          'Cranberry Muffin',
          'Glazed Blueberry Cake Donut',
          'Glazed Marble Cake Donut',
          'Pecan Honey Bun' ],
        'Breakfast Misc': 
         [ 'Pineapple',
          'Strawberries',
          'Vegan Grits Hot Cereal',
          'Watermelon' ] }
   (end)

    */
   function getMenu (loc, meal, cb) {
      var locId = locations[loc].id,
          url;

      meal = ucwords(meal);
      url = baseURL + "LocationNUM=" + locId + "&mealName=" + meal;

      request(url, function(err, response, data) {
         var menu, category, items, i, food_name;
         if(err) {
            return cb(err, null);
         }

         xmlparse(data, function(err, dom) {
            if(err) {
               return cb(err, null);
            }

            menu = {};
            // make the 'default' category, see below for more info
            category = "Menu";
            menu[category] = []; 
            items = dom.document.getElementsByTagName('item');

            for(i = 0; i < items.length; i++) {
               food_name = items[i].childNodes[0].innerHTML;
               // we need to change the case of the names and add them to the 
               // object as expected.
               if(isCategory(food_name)) {
                  category = fixCategoryName(food_name);
                  menu[category] = [];
               }

               else {
                  menu[category].push(fixFood(food_name));
               }
            }

            // Some dining halls (Livingston Dining Commons) bizarrely do not
            // have category names, so we have the "Menu" category for these
            // situations.  If it's empty, then we do have category names, so
            // we'll delete it.

            if (menu.Menu.length === 0) {
               delete menu.Menu;
            }

            cb(null, menu);
         });

      });
   }

   /* Function: getLocations
    * Gets the list of dining locations.
    */

   function getLocations () {
      return locations;
   }

   /* 
    Group: Private functions
   
    Function: isCategory
    Takes in the name of a food item (returned by parsing the XML feed) and 
    tells you if it is the name of a category.
   
    Parameters:
      food_name - *string* The name of the food item.
    
    Returns:
      true if the food item is a category name, false otherwise.
   
    Example:
      isCategory('broccoli') = false
      isCategory('-- broccoli --') = true
    */

   function isCategory (food_name) {
      return food_name.match(/-- .+ --/);
   }

   /* 
    Function: ucwords
    Capitalizes the first letter of every word in the input string, and 
    lowercases everything else. We use this to fix the shitty meal data 
    we get from the Rutgers Dining Hall APIs.
   
    Parameters:
      str - *string*
    
    Returns:
      A new string with the first character of every word capitalized, and the rest of the word in lowercase.
    
    Example:
      ucwords("HELLO WORLD") = "Hello World"
    */

   function ucwords (str) {
      return (str + '').replace(/[A-Za-z]+/g, function (word) {
         return word.charAt(0).toUpperCase() + word.substring(1).toLowerCase();
      });
   }

   /* 
    Function: fixCategoryName
    Removes the "--"s around the name of a category obtained from a RUDining 
    XML Feed to give us the name of the category itself capitalized properly.

    Parameters:
      name - The name of a category with --s around it.
   
    Return:
      A new string without the surrounding --s with every word capitalized correctly.
    
    Example:
      fixCategoryName("-- BREAKFAST MEATS --") // => "Breakfast Meats"
    */

   function fixCategoryName (name) {
      name = name.replace(/-- (.+) --/, '$1');
      name = name.replace('&amp;', '&');
      name = ucwords(name);
      name = name.replace(/ +/g, ' ');
      return name;
   }

   /* Function: fixFood
    * Cleans up a food name entry, fixing ampersands and such
    *
    * Parameters:
    *    food     - *string* food name
    *
    * Returns:
    *    *string* cleaner food name
    */

   function fixFood (food) {
      food = food.replace('&amp;', '&');
      food = ucwords(food);
      food = food.replace(/ +/g, ' ');
      return food;
   }

   exports.getMenu = getMenu;
   exports.getLocations = getLocations;

   return exports;
};

